#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import pytest
import uuid

from dateutil.tz import tzutc

from polyaxon.exceptions import PolyaxonSchemaError
from polyaxon.parser import parser
from polyaxon.parser.constants import NO_VALUE_FOUND
from polyaxon.schemas.types import (
    V1ArtifactsType,
    V1AuthType,
    V1DockerfileType,
    V1FileType,
    V1GcsType,
    V1GitType,
    V1S3Type,
    V1UriType,
    V1WasbType,
)
from tests.utils import BaseTestCase


@pytest.mark.parser_mark
class TestParser(BaseTestCase):
    def test_get_boolean(self):
        value = parser.get_boolean(key="bool_key_1", value="1")
        self.assertEqual(value, True)

        value = parser.get_boolean(key="bool_key_2", value="true")
        self.assertEqual(value, True)

        value = parser.get_boolean(key="bool_key_2", value=True)
        self.assertEqual(value, True)

        value = parser.get_boolean(key="bool_key_3", value="0")
        self.assertEqual(value, False)

        value = parser.get_boolean(key="bool_key_4", value="false")
        self.assertEqual(value, False)

        value = parser.get_boolean(key="bool_key_4", value=False)
        self.assertEqual(value, False)

        value = parser.get_boolean(
            key="bool_list_key_1",
            value=[False, "false", True, "true", "1", "0"],
            is_list=True,
        )
        self.assertEqual(value, [False, False, True, True, True, False])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_error_key_1", value="null")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_error_key_1", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_error_key_2", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_error_key_3", value=0)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_error_key_4", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_error_key_5", value="")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(
                key="bool_list_key_1", value=[False, "false", True, "true", "1", "0"]
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(
                key="bool_list_error_key_2", value=[False, 1, 0], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(
                key="bool_list_error_key_1",
                value=[False, "false", True, "true", "1", "0", "foo"],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_key_1", value="1", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_key_2", value=True, is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_boolean(key="bool_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_boolean(
                key="bool_non_existing_key", value=None, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_boolean(
                key="bool_non_existing_key", value=None, is_optional=True, default=True
            ),
            True,
        )

        self.assertEqual(
            parser.get_boolean(
                key="bool_non_existing_key", value=None, is_list=True, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_boolean(
                key="bool_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=[True, False],
            ),
            [True, False],
        )

    def test_get_int(self):
        value = parser.get_int(key="int_key_1", value=123)
        self.assertEqual(value, 123)

        value = parser.get_int(key="int_key_2", value="123")
        self.assertEqual(value, 123)

        value = parser.get_int(
            key="int_list_key_1", value=["123", 124, 125, "125"], is_list=True
        )
        self.assertEqual(value, [123, 124, 125, 125])

        value = parser.get_int(
            key="int_list_key_1", value='["123", 124, 125, "125"]', is_list=True
        )
        self.assertEqual(value, [123, 124, 125, 125])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_error_key_1", value="null")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_error_key_1", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_error_key_2", value="")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_error_key_3", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_list_key_1", value=["123", 124, 125, "125"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(
                key="int_list_error_key_1",
                value=["123", 124, 125, "125", None],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(
                key="int_list_error_key_2",
                value=["123", 1.24, 125, "125"],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(
                key="int_list_error_key_3",
                value=["123", 1.24, 125, "foo"],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_key_1", value=125, is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_key_2", value="125", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_int(key="int_non_existing_key", value=NO_VALUE_FOUND)

        self.assertEqual(
            parser.get_int(key="int_non_existing_key", value=None, is_optional=True),
            None,
        )
        self.assertEqual(
            parser.get_int(
                key="int_non_existing_key", value=None, is_optional=True, default=34
            ),
            34,
        )

        self.assertEqual(
            parser.get_int(
                key="int_non_existing_key", value=None, is_list=True, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_int(
                key="int_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=[34, 1],
            ),
            [34, 1],
        )

    def test_get_float(self):
        value = parser.get_float(key="float_key_1", value=1.23)
        self.assertEqual(value, 1.23)

        value = parser.get_float(key="float_key_2", value="1.23")
        self.assertEqual(value, 1.23)

        value = parser.get_float(key="float_key_3", value="123")
        self.assertEqual(value, 123)

        value = parser.get_float(
            key="float_list_key_1", value=[1.23, 13.3, "4.4", "555", 66.0], is_list=True
        )
        self.assertEqual(value, [1.23, 13.3, 4.4, 555.0, 66.0])

        value = parser.get_float(
            key="float_list_key_1",
            value='[1.23, 13.3, "4.4", "555", 66.0]',
            is_list=True,
        )
        self.assertEqual(value, [1.23, 13.3, 4.4, 555.0, 66.0])

        value = parser.get_float(key="float_from_int", value=123)
        self.assertEqual(value, 123.0)

        value = parser.get_float(
            key="float_key_2", value=[1.23, 13.3, 66], is_list=True
        )
        self.assertEqual(value, [1.23, 13.3, 66])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_error_key_1", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_error_key_1", value="null")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_error_key_2", value="")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_error_key_3", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(
                key="float_list_key_1", value=[1.23, 13.3, "4.4", "555", 66.0]
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_list_error_key_1", value=None, is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_list_error_key_2", value="", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_list_error_key_3", value="foo", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_key_1", value=213, is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(key="float_non_existing_key", value=[1.23, 13.3, "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_float(
                key="float_non_existing_key", value=[1.23, 13.3, None], is_list=True
            )

        self.assertEqual(
            parser.get_float(
                key="float_non_existing_key", value=None, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_float(
                key="float_non_existing_key", value=None, is_optional=True, default=3.4
            ),
            3.4,
        )

        self.assertEqual(
            parser.get_float(
                key="float_non_existing_key",
                value="null",
                is_list=True,
                is_optional=True,
            ),
            None,
        )
        self.assertEqual(
            parser.get_float(
                key="float_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=[3.4, 1.2],
            ),
            [3.4, 1.2],
        )

    def test_get_string(self):
        value = parser.get_string(key="string_key_1", value="123")
        self.assertEqual(value, "123")

        value = parser.get_string(key="string_key_2", value="1.23")
        self.assertEqual(value, "1.23")

        value = parser.get_string(key="string_key_3", value="foo")
        self.assertEqual(value, "foo")

        value = parser.get_string(key="string_key_4", value="")
        self.assertEqual(value, "")

        value = parser.get_string(
            key="string_list_key_1", value=["123", "1.23", "foo", ""], is_list=True
        )
        self.assertEqual(value, ["123", "1.23", "foo", ""])

        value = parser.get_string(
            key="string_list_key_1", value='["123", "1.23", "foo", ""]', is_list=True
        )
        self.assertEqual(value, ["123", "1.23", "foo", ""])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_error_key_1", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_error_key_2", value=123)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_error_key_3", value=1.23)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_error_key_4", value=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_list_key_1", value=["123", "1.23", "foo", ""])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(
                key="string_list_error_key_1", value=["123", 123], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(
                key="string_list_error_key_2", value=["123", 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(
                key="string_list_error_key_3", value=["123", None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(
                key="string_list_error_key_4", value=["123", False], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_key_3", value="foo", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_key_4", value="", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_string(key="string_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_string(
                key="string_non_existing_key", value=None, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_string(
                key="string_non_existing_key",
                value=None,
                is_optional=True,
                default="foo",
            ),
            "foo",
        )

        self.assertEqual(
            parser.get_string(
                key="string_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
            ),
            None,
        )
        self.assertEqual(
            parser.get_string(
                key="string_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=["foo", "bar"],
            ),
            ["foo", "bar"],
        )

    def test_get_dict(self):
        value = parser.get_dict(
            key="dict_key_1",
            value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
        )
        self.assertEqual(value, {"key1": "foo", "key2": 2, "key3": False, "key4": "1"})

        value = parser.get_dict(
            key="dict_key_1",
            value='{"key1": "foo", "key2": 2, "key3": false, "key4": "1"}',
        )
        self.assertEqual(value, {"key1": "foo", "key2": 2, "key3": False, "key4": "1"})

        value = parser.get_dict(
            key="dict_key_1",
            value='{"key1": "foo", "key2": 2, "key3": false, "key4": "1"}',
        )
        self.assertEqual(value, {"key1": "foo", "key2": 2, "key3": False, "key4": "1"})

        value = parser.get_dict(
            key="dict_list_key_1",
            value=[
                {"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                {"key3": True, "key4": "2"},
                {"key1": False, "key2": "3"},
            ],
            is_list=True,
        )
        self.assertEqual(
            value,
            [
                {"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                {"key3": True, "key4": "2"},
                {"key1": False, "key2": "3"},
            ],
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_list_key_1", value=["123", {"key3": True}])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(
                key="dict_list_error_key_1", value=["123", {"key3": True}], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(
                key="dict_list_error_key_2", value=[{"key3": True}, 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(
                key="dict_list_error_key_3", value=[{"key3": True}, None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(
                key="dict_list_error_key_4",
                value=[{"key3": True}, "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(
                key="dict_key_1",
                value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict(key="dict_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_dict(key="dict_non_existing_key", value=None, is_optional=True),
            None,
        )
        self.assertEqual(
            parser.get_dict(
                key="dict_non_existing_key",
                value=None,
                is_optional=True,
                default={"foo": "bar"},
            ),
            {"foo": "bar"},
        )

        self.assertEqual(
            parser.get_dict(
                key="dict_non_existing_key", value=None, is_list=True, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_dict(
                key="dict_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=[{"foo": "bar"}, {"foo": "boo"}],
            ),
            [{"foo": "bar"}, {"foo": "boo"}],
        )

    def test_get_uri(self):
        value = parser.get_uri(key="uri_key_1", value="user:pass@siteweb.ca")
        self.assertEqual(value, V1UriType("user", "pass", "siteweb.ca"))

        value = parser.get_uri(key="uri_key_2", value="user2:pass@localhost:8080")
        self.assertEqual(value, V1UriType("user2", "pass", "localhost:8080"))

        value = parser.get_uri(key="uri_key_3", value="user2:pass@https://quay.io")
        self.assertEqual(value, V1UriType("user2", "pass", "https://quay.io"))

        value = parser.get_uri(
            key="uri_list_key_1",
            value=[
                "user:pass@siteweb.ca",
                "user2:pass@localhost:8080",
                "user2:pass@https://quay.io",
            ],
            is_list=True,
        )
        self.assertEqual(
            value,
            [
                V1UriType("user", "pass", "siteweb.ca"),
                V1UriType("user2", "pass", "localhost:8080"),
                V1UriType("user2", "pass", "https://quay.io"),
            ],
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(
                key="uri_list_key_1",
                value=[
                    "user:pass@siteweb.ca",
                    "user2:pass@localhost:8080",
                    "user2:pass@https://quay.io",
                ],
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(
                key="uri_list_error_key_1",
                value=["123", "user:pass@siteweb.ca"],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(
                key="uri_list_error_key_2",
                value=["user:pass@siteweb.ca", 12.3],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(
                key="uri_list_error_key_3",
                value=["user:pass@siteweb.ca", None],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(
                key="uri_list_error_key_4",
                value=["user:pass@siteweb.ca", "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_key_1", value="user:pass@siteweb.ca", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uri(key="uri_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_uri(key="uri_non_existing_key", value=None, is_optional=True),
            None,
        )
        self.assertEqual(
            parser.get_uri(
                key="uri_non_existing_key",
                value=None,
                is_optional=True,
                default=V1UriType("user2", "pass", "localhost:8080"),
            ),
            V1UriType("user2", "pass", "localhost:8080"),
        )

        self.assertEqual(
            parser.get_uri(
                key="uri_non_existing_key", value=None, is_list=True, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_uri(
                key="uri_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=[
                    V1UriType("user", "pass", "siteweb.ca"),
                    V1UriType("user2", "pass", "localhost:8080"),
                ],
            ),
            [
                V1UriType("user", "pass", "siteweb.ca"),
                V1UriType("user2", "pass", "localhost:8080"),
            ],
        )

    def test_get_auth(self):
        value = parser.get_auth(
            key="auth_key_1", value={"user": "user", "password": "pass"}
        )
        self.assertEqual(value, V1AuthType("user", "pass"))

        value = parser.get_auth(key="auth_key_1", value=V1AuthType("user", "pass"))
        self.assertEqual(value, V1AuthType("user", "pass"))

        value = parser.get_auth(key="auth_key_1", value="user:pass")
        self.assertEqual(value, V1AuthType("user", "pass"))

        value = parser.get_auth(
            key="auth_list_key_1", value=["user:pass", "user2:pass"], is_list=True
        )
        self.assertEqual(
            value, [V1AuthType("user", "pass"), V1AuthType("user2", "pass")]
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_list_key_1", value=["user:pass", "user2:pass"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(
                key="auth_list_error_key_1", value=["123", "user:pass"], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(
                key="auth_list_error_key_2", value=["user:pass", 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(
                key="auth_list_error_key_3", value=["user:pass", None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(
                key="auth_list_error_key_4",
                value=["user:pass", "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_key_1", value="user:pass", is_list=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_auth(key="auth_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_auth(key="auth_non_existing_key", value=None, is_optional=True),
            None,
        )
        self.assertEqual(
            parser.get_auth(
                key="auth_non_existing_key",
                value=None,
                is_optional=True,
                default=V1AuthType("user2", "pass"),
            ),
            V1AuthType("user2", "pass"),
        )

        self.assertEqual(
            parser.get_auth(
                key="auth_non_existing_key", value=None, is_list=True, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_auth(
                key="auth_non_existing_key",
                value=None,
                is_list=True,
                is_optional=True,
                default=[V1AuthType("user", "pass"), V1AuthType("user2", "pass")],
            ),
            [V1AuthType("user", "pass"), V1AuthType("user2", "pass")],
        )

    def test_get_list(self):
        value = parser.get_list(
            key="list_key_1", value="user:pass@siteweb.ca, 'pp', 0.1, 'foo'"
        )
        self.assertEqual(value, ["user:pass@siteweb.ca", "'pp'", "0.1", "'foo'"])

        value = parser.get_list(
            key="list_key_2", value="user1,user2 , user3,     user4    , user5"
        )
        self.assertEqual(value, ["user1", "user2", "user3", "user4", "user5"])

        value = parser.get_list(key="list_key_3", value=[False])
        self.assertEqual(value, [False])

        value = parser.get_list(key="list_key_3", value=["false"])
        self.assertEqual(value, ["false"])

        value = parser.get_list(key="list_key_4", value="foo")
        self.assertEqual(value, ["foo"])

        value = parser.get_list(key="list_key_5", value="")
        self.assertEqual(value, [])

        value = parser.get_list(key="list_error_key_3", value="null")
        self.assertEqual(value, ["null"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_list(key="list_error_key_1", value=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_list(key="list_error_key_2", value={"key": "value"})

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_list(key="list_error_key_4", value=123)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_list(key="list_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_list(key="list_non_existing_key", value=NO_VALUE_FOUND)

        self.assertEqual(
            parser.get_list(key="list_non_existing_key", value=None, is_optional=True),
            None,
        )
        self.assertEqual(
            parser.get_list(
                key="list_non_existing_key",
                value=None,
                is_optional=True,
                default=["foo"],
            ),
            ["foo"],
        )

    def test_get_dict_of_dicts(self):
        value = parser.get_dict_of_dicts(
            key="dict_dicts_key_1",
            value={"data1": {"mountPath": "/data/21", "existingClaim": "data-1-pvc"}},
        )
        self.assertEqual(
            value, {"data1": {"mountPath": "/data/21", "existingClaim": "data-1-pvc"}}
        )

        value = parser.get_dict_of_dicts(
            key="dict_dicts_key_1",
            value='{"data1": {"mountPath": "/data/21", "existingClaim": "data-1-pvc"}}',
        )
        self.assertEqual(
            value, {"data1": {"mountPath": "/data/21", "existingClaim": "data-1-pvc"}}
        )

        value = parser.get_dict_of_dicts(
            key="dict_dicts_key_2",
            value={
                "outputs1": {
                    "mountPath": "/output/2",
                    "existingClaim": "outputs-1-pvc",
                },
                "outputs2": {"mountPath": "/output/2", "existingClaim": "output-2-pvc"},
            },
        )
        self.assertEqual(
            value,
            {
                "outputs1": {
                    "mountPath": "/output/2",
                    "existingClaim": "outputs-1-pvc",
                },
                "outputs2": {"mountPath": "/output/2", "existingClaim": "output-2-pvc"},
            },
        )

        value = parser.get_dict_of_dicts(key="dict_dicts_key_3", value={})
        self.assertEqual(value, None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(key="dict_dicts_error_key_1", value=True)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(
                key="dict_dicts_error_key_2", value={"key": "value"}
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(key="dict_dicts_error_key_3", value="null")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(key="dict_dicts_error_key_4", value=123)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(key="dict_dicts_error_key_5", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(key="dict_dicts_error_key_6", value="")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(
                key="dict_dicts_error_key_7",
                value={"mountPath": "/data/2", "existingClaim": "data-2-pvc"},
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(key="dict_dicts_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dict_of_dicts(
                key="dict_dicts_non_existing_key", value=NO_VALUE_FOUND
            )

        self.assertEqual(
            parser.get_dict_of_dicts(
                key="dict_dicts_non_existing_key", value=None, is_optional=True
            ),
            None,
        )
        self.assertEqual(
            parser.get_dict_of_dicts(
                key="dict_dicts_non_existing_key",
                value=None,
                is_optional=True,
                default={},
            ),
            {},
        )

    def test_get_wasbs_path(self):
        # Correct url
        wasbs_path = "wasbs://container@user.blob.core.windows.net/path"
        expected = V1WasbType("container", "user", "path")
        parsed_url = parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == expected
        parsed_url = parser.get_wasbs_path(key="wasb_key", value=wasbs_path)
        assert parsed_url == expected

        wasbs_path = "wasbs://container@user.blob.core.windows.net/"
        expected = V1WasbType("container", "user", "")
        parsed_url = parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == expected
        parsed_url = parser.get_wasbs_path(key="wasb_key", value=wasbs_path)
        assert parsed_url == expected

        wasbs_path = "wasbs://container@user.blob.core.windows.net"
        expected = V1WasbType("container", "user", "")
        parsed_url = parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == expected
        parsed_url = parser.get_wasbs_path(key="wasb_key", value=wasbs_path)
        assert parsed_url == expected

        wasbs_path = "wasbs://container@user.blob.core.windows.net/path/to/file"
        expected = V1WasbType("container", "user", "path/to/file")
        parsed_url = parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == expected
        parsed_url = parser.get_wasbs_path(key="wasb_key", value=wasbs_path)
        assert parsed_url == expected

        # Wrong url
        wasbs_path = "wasbs://container@user.foo.bar.windows.net/path/to/file"
        with self.assertRaises(PolyaxonSchemaError):
            parser.parse_wasbs_path(wasbs_path)
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_wasbs_path(key="wasb_key", value=wasbs_path)

        wasbs_path = "wasbs://container@user.blob.core.foo.net/path/to/file"
        with self.assertRaises(PolyaxonSchemaError):
            parser.parse_wasbs_path(wasbs_path)
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_wasbs_path(key="wasb_key", value=wasbs_path)

        wasbs_path = "wasbs://container@user.blob.windows.net/path/to/file"
        with self.assertRaises(PolyaxonSchemaError):
            parser.parse_wasbs_path(wasbs_path)
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_wasbs_path(key="wasb_key", value=wasbs_path)

    def test_parse_gcs_path(self):
        # Correct url
        gcs_path = "gs://bucket/path/to/blob"
        expected = V1GcsType("bucket", "path/to/blob")
        parsed_url = parser.parse_gcs_path(gcs_path)
        assert parsed_url == expected
        parsed_url = parser.get_gcs_path(key="gcs_key", value=gcs_path)
        assert parsed_url == expected

        # Wrong url
        gcs_path = "gs:/bucket/path/to/blob"
        with self.assertRaises(PolyaxonSchemaError):
            parser.parse_gcs_path(gcs_path)
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_gcs_path(key="gcs_key", value=gcs_path)

        # Trailing slash
        gcs_path = "gs://bucket/path/to/blob/"
        expected = V1GcsType("bucket", "path/to/blob/")
        assert parser.parse_gcs_path(gcs_path) == expected
        parsed_url = parser.get_gcs_path(key="gcs_key", value=gcs_path)
        assert parsed_url == expected

        # Bucket only
        gcs_path = "gs://bucket/"
        expected = V1GcsType("bucket", "")
        assert parser.parse_gcs_path(gcs_path) == expected
        parsed_url = parser.get_gcs_path(key="gcs_key", value=gcs_path)
        assert parsed_url == expected

    def test_parse_s3_path(self):
        s3_path = "s3://test/this/is/bad/key.txt"
        expected = V1S3Type("test", "this/is/bad/key.txt")
        parsed_url = parser.parse_s3_path(s3_path)
        assert parsed_url == expected
        parsed_url = parser.get_s3_path(key="s3_key", value=s3_path)
        assert parsed_url == expected

    def test_parse_date(self):
        value = "2010-12-12"
        parsed_url = parser.get_date(key="date_key", value=value)
        assert parsed_url == datetime.date(2010, 12, 12)

        value = datetime.date(2010, 12, 12)
        parsed_url = parser.get_date(key="date_key", value=value)
        assert parsed_url == value

        value = "2010-12-12-12"
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_date(key="date_key", value=value)

    def test_parse_datetime(self):
        value = "2010-12-12 10:10"
        parsed_url = parser.get_datetime(key="date_key", value=value)
        assert parsed_url == datetime.datetime(2010, 12, 12, 10, 10)

        value = "2010-12-12 01:00"
        parsed_url = parser.get_datetime(key="date_key", value=value)
        assert parsed_url == datetime.datetime(2010, 12, 12, 1, 0)

        value = "2010-12-12 01:53:12"
        parsed_url = parser.get_datetime(key="date_key", value=value)
        assert parsed_url == datetime.datetime(2010, 12, 12, 1, 53, 12)

        value = "2014-12-22T03:12:58.019077+00:00"
        parsed_url = parser.get_datetime(key="date_key", value=value)
        assert parsed_url == datetime.datetime(
            2014, 12, 22, 3, 12, 58, 19077, tzinfo=tzutc()
        )

        value = datetime.datetime(2010, 12, 12, 0, 0, 0)
        parsed_url = parser.get_datetime(key="date_key", value=value)
        assert parsed_url == value

        value = datetime.datetime(2010, 12, 12, 0, 0, 0, tzinfo=tzutc())
        parsed_url = parser.get_datetime(key="date_key", value=value)
        assert parsed_url == value

        # Dates are not validate by datetime
        value = "2010-12-12"
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_datetime(key="date_key", value=value)

    def test_parse_uuid(self):
        value = uuid.uuid4()
        parsed_uid = parser.get_uuid(key="uuid_key", value=value)
        assert parsed_uid == value.hex

        parsed_uid = parser.get_uuid(key="uuid_key", value=value.hex)
        assert parsed_uid == value.hex

        value = "2sd2"
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uuid(key="uuid_key", value=value)

        value = "2sd2-sdf"
        with self.assertRaises(PolyaxonSchemaError):
            parser.get_uuid(key="uuid_key", value=value)

    def test_get_dockerfile_init(self):
        value = parser.get_dockerfile_init(
            key="dict_key_1", value={"image": "foo", "env": {"key1": 2, "key2": 21}}
        )
        self.assertEqual(
            value, V1DockerfileType(image="foo", env={"key1": 2, "key2": 21})
        )

        value = parser.get_dockerfile_init(
            key="dict_key_1", value='{"image": "foo", "env": {"key1": 2, "key2": 21}}'
        )
        self.assertEqual(
            value, V1DockerfileType(image="foo", env={"key1": 2, "key2": 21})
        )

        value = parser.get_dockerfile_init(
            key="dict_key_1", value='{"image": "foo", "run": ["exec1", "exec2"]}'
        )
        self.assertEqual(value, V1DockerfileType(image="foo", run=["exec1", "exec2"]))

        value = parser.get_dockerfile_init(
            key="dict_list_key_1",
            value=[
                {"image": "foo", "env": {"key1": 2, "key2": 21}},
                {"image": "foo2", "copy": ["exec1", "exec2"]},
                {"image": "foo3", "run": ["exec1", "exec2"]},
            ],
            is_list=True,
        )
        self.assertEqual(
            value,
            [
                V1DockerfileType(image="foo", env={"key1": 2, "key2": 21}),
                V1DockerfileType(image="foo2", copy=["exec1", "exec2"]),
                V1DockerfileType(image="foo3", run=["exec1", "exec2"]),
            ],
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(key="dict_error_key_1", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(key="dict_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(key="dict_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(key="dict_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(key="dict_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_list_key_1", value=["123", {"key3": True}]
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_list_error_key_1", value=["123", {"key3": True}], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_list_error_key_2", value=[{"key3": True}, 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_list_error_key_3", value=[{"key3": True}, None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_list_error_key_4",
                value=[{"key3": True}, "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_key_1",
                value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(key="dict_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_non_existing_key", value=NO_VALUE_FOUND
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_dockerfile_init(
                key="dict_non_existing_key", value=None, is_list=True
            )

        self.assertEqual(
            parser.get_dockerfile_init(
                key="dict_non_existing_key", value=None, is_optional=True
            ),
            None,
        )

    def test_get_file_init(self):
        value = parser.get_file_init(
            key="dict_key_1", value={"filename": "foo.yaml", "content": "test"}
        )
        self.assertEqual(value, V1FileType(filename="foo.yaml", content="test"))

        value = parser.get_file_init(key="dict_key_1", value={"content": "test"})
        self.assertEqual(value, V1FileType(content="test"))

        value = parser.get_file_init(
            key="dict_key_1", value='{"filename": "foo.yaml", "content": "test"}'
        )
        self.assertEqual(value, V1FileType(filename="foo.yaml", content="test"))

        value = parser.get_file_init(key="dict_key_1", value='{"content": "test"}')
        self.assertEqual(value, V1FileType(content="test"))

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_error_key_1",
                value=dict(content="foo", connection="foo", init=True),
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_error_key_1", value=dict(content="foo", init=True)
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_error_key_1", value=dict(content="foo", connection="foo")
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_list_key_1", value=["123", {"key3": True}])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_list_error_key_1", value=["123", {"key3": True}], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_list_error_key_2", value=[{"key3": True}, 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_list_error_key_3", value=[{"key3": True}, None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_list_error_key_4",
                value=[{"key3": True}, "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(
                key="dict_key_1",
                value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_file_init(key="dict_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_file_init(
                key="dict_non_existing_key", value=None, is_optional=True
            ),
            None,
        )

    def test_get_git_init(self):
        value = parser.get_git_init(key="dict_key_1", value={"revision": "foo"})
        self.assertEqual(value, V1GitType(revision="foo"))

        value = parser.get_git_init(
            key="dict_key_1",
            value={"revision": "foo"},
        )
        self.assertEqual(value, V1GitType(revision="foo"))

        value = parser.get_git_init(
            key="dict_key_1", value={"url": "https://github.com", "revision": "foo"}
        )
        self.assertEqual(value, V1GitType(revision="foo", url="https://github.com"))

        value = parser.get_git_init(
            key="dict_key_1", value='{"revision": "foo", "url": "https://github.com"}'
        )
        self.assertEqual(value, V1GitType(revision="foo", url="https://github.com"))

        value = parser.get_git_init(key="dict_key_1", value='{"revision": "foo"}')
        self.assertEqual(value, V1GitType(revision="foo"))

        value = parser.get_git_init(
            key="dict_list_key_1",
            value=[
                {"revision": "foo"},
                {"url": "https://github.com", "revision": "foo"},
                {"url": "https://github.com"},
            ],
            is_list=True,
        )
        self.assertEqual(
            value,
            [
                V1GitType(revision="foo"),
                V1GitType(revision="foo", url="https://github.com"),
                V1GitType(url="https://github.com"),
            ],
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_error_key_1",
                value=dict(revision="foo", connection="foo", init=True),
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_error_key_1", value=dict(revision="foo", init=True)
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_error_key_1", value=dict(revision="foo", connection="foo")
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_list_key_1", value=["123", {"key3": True}])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_list_error_key_1", value=["123", {"key3": True}], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_list_error_key_2", value=[{"key3": True}, 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_list_error_key_3", value=[{"key3": True}, None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_list_error_key_4",
                value=[{"key3": True}, "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(
                key="dict_key_1",
                value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_git_init(key="dict_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_git_init(
                key="dict_non_existing_key", value=None, is_optional=True
            ),
            None,
        )

    def test_get_image_init(self):
        value = parser.get_image_init(key="dict_key_1", value="foo")
        self.assertEqual(value, "foo")

        value = parser.get_image_init(key="dict_key_1", value={"name": "foo"})
        self.assertEqual(value, "foo")

        value = parser.get_image_init(key="dict_key_1", value="foo:bar")
        self.assertEqual(value, "foo:bar")

        value = parser.get_image_init(
            key="dict_key_1", value={"name": "foo:bar", "connection": "foo"}
        )
        self.assertEqual(value, "foo:bar")

        value = parser.get_image_init(
            key="dict_key_1", value="https://registry.com/foo:bar"
        )
        self.assertEqual(value, "https://registry.com/foo:bar")

        value = parser.get_image_init(
            key="dict_key_1", value='{"name": "https://registry.com/foo:bar"}'
        )
        self.assertEqual(value, "https://registry.com/foo:bar")

        value = parser.get_image_init(
            key="dict_list_key_1",
            value=[
                {"name": "https://registry.com/foo:bar"},
                {"name": "test", "connection": "registry"},
                "foo:bar",
            ],
            is_list=True,
        )
        self.assertEqual(
            value,
            [
                "https://registry.com/foo:bar",
                "test",
                "foo:bar",
            ],
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_list_key_1", value=["123", {"key3": True}])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(
                key="dict_list_error_key_1", value=["123", {"key3": True}], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(
                key="dict_list_error_key_2", value=[{"key3": True}, 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(
                key="dict_list_error_key_3", value=[{"key3": True}, None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(
                key="dict_list_error_key_4",
                value=[{"key3": True}, "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(
                key="dict_key_1",
                value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_image_init(key="dict_non_existing_key", value=None, is_list=True)

        self.assertEqual(
            parser.get_image_init(
                key="dict_non_existing_key", value=None, is_optional=True
            ),
            None,
        )

    def test_get_artifacts_init(self):
        value = parser.get_artifacts_init(
            key="dict_key_1", value={"files": ["foo", "bar"]}
        )
        self.assertEqual(value, V1ArtifactsType(files=["foo", "bar"]))

        value = parser.get_artifacts_init(
            key="dict_key_1", value={"files": [["from-foo", "to-foo"], "bar"]}
        )
        self.assertEqual(value, V1ArtifactsType(files=[["from-foo", "to-foo"], "bar"]))

        value = parser.get_artifacts_init(
            key="dict_key_1", value='{"dirs": ["foo", "bar"]}'
        )
        self.assertEqual(value, V1ArtifactsType(dirs=["foo", "bar"]))

        value = parser.get_artifacts_init(
            key="dict_key_1", value='{"dirs": [["from-foo", "to-foo"], "bar"]}'
        )
        self.assertEqual(value, V1ArtifactsType(dirs=[["from-foo", "to-foo"], "bar"]))

        value = parser.get_artifacts_init(
            key="dict_list_key_1",
            value=[
                {
                    "dirs": [["from-foo", "to-foo"], "bar"],
                    "files": [["from-foo", "to-foo"], "bar"],
                },
                {"files": [["from-foo", "to-foo"], "bar"]},
            ],
            is_list=True,
        )
        self.assertEqual(
            value,
            [
                V1ArtifactsType(
                    dirs=[["from-foo", "to-foo"], "bar"],
                    files=[["from-foo", "to-foo"], "bar"],
                ),
                V1ArtifactsType(files=[["from-foo", "to-foo"], "bar"]),
            ],
        )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_key_1", value={"connection": "foo", "init": True}
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_key_1", value={"init": True, "dirs": ["foo", "bar"]}
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_key_1", value={"connection": "foo", "files": ["foo", "bar"]}
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_error_key_1", value={"paths": ["foo", "bar"]}
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(key="dict_error_key_1", value="foo")

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(key="dict_error_key_2", value=1)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(key="dict_error_key_3", value=False)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(key="dict_error_key_4", value=["1", "foo"])

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_list_key_1", value=["123", {"key3": True}]
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_list_error_key_1", value=["123", {"key3": True}], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_list_error_key_2", value=[{"key3": True}, 12.3], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_list_error_key_3", value=[{"key3": True}, None], is_list=True
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_list_error_key_4",
                value=[{"key3": True}, "123", False],
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_key_1",
                value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                is_list=True,
            )

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(key="dict_non_existing_key", value=None)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(key="dict_non_existing_key", value=NO_VALUE_FOUND)

        with self.assertRaises(PolyaxonSchemaError):
            parser.get_artifacts_init(
                key="dict_non_existing_key", value=None, is_list=True
            )

        self.assertEqual(
            parser.get_artifacts_init(
                key="dict_non_existing_key", value=None, is_optional=True
            ),
            None,
        )



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from polyaxon.exceptions import PolypodException
from polyaxon.k8s import k8s_schemas
from polyaxon.polyflow.environment import V1Environment
from polyaxon.polypod.pod.spec import get_pod_spec
from tests.utils import BaseTestCase


@pytest.mark.polypod_mark
class TestPodSpec(BaseTestCase):
    def test_get_pod_spec(self):
        init_container = k8s_schemas.V1Container(name="init")
        main_container = k8s_schemas.V1Container(name="main")
        sidecar_container = k8s_schemas.V1Container(name="sidecar")
        volumes = [k8s_schemas.V1Volume(name="vol")]
        labels = {"key": "labels"}
        annotations = {"key": "annotations"}
        node_selector = {"key": "selector"}
        affinity = [{"key": "affinity"}]
        tolerations = {"key": "tolerations"}
        security_context = {"uid": 222, "gid": 222}
        restart_policy = "never"

        with self.assertRaises(PolypodException):
            get_pod_spec(
                namespace="default",
                main_container=None,
                sidecar_containers=None,
                init_containers=None,
                resource_name="foo",
                volumes=None,
                environment=V1Environment(),
                labels={},
            )

        environment = V1Environment(
            service_account_name="sa",
            labels=labels,
            annotations=annotations,
            node_selector=node_selector,
            affinity=affinity,
            tolerations=tolerations,
            security_context=security_context,
            image_pull_secrets=[],
            restart_policy=restart_policy,
        )
        metadata, pod_spec = get_pod_spec(
            namespace="default",
            resource_name="foo",
            main_container=main_container,
            sidecar_containers=None,
            init_containers=None,
            volumes=None,
            environment=environment,
            labels=environment.labels,
        )

        assert metadata.name == "foo"
        assert metadata.labels == labels
        assert metadata.namespace == "default"
        assert metadata.annotations == annotations

        assert isinstance(pod_spec, k8s_schemas.V1PodSpec)
        assert pod_spec.security_context == security_context
        assert pod_spec.restart_policy == "never"
        assert pod_spec.service_account_name == "sa"
        assert pod_spec.init_containers == []
        assert pod_spec.containers == [main_container]
        assert pod_spec.volumes is None
        assert pod_spec.node_selector == node_selector
        assert pod_spec.tolerations == tolerations
        assert pod_spec.affinity == affinity

        environment = V1Environment(
            service_account_name="sa",
            labels=labels,
            annotations=annotations,
            node_selector=node_selector,
            affinity=affinity,
            tolerations=tolerations,
            security_context=security_context,
            image_pull_secrets=[],
            restart_policy=restart_policy,
        )
        metadata, pod_spec = get_pod_spec(
            namespace="default",
            main_container=main_container,
            sidecar_containers=[sidecar_container],
            init_containers=[init_container],
            resource_name="foo",
            volumes=volumes,
            environment=environment,
            labels={},
        )

        assert pod_spec.init_containers == [init_container]
        assert pod_spec.containers == [main_container, sidecar_container]
        assert pod_spec.volumes == volumes
        assert metadata.labels == {}



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import uuid

from mock import MagicMock, mock

from polyaxon import settings
from polyaxon.client import RunClient
from polyaxon.lifecycle import V1Statuses
from tests.utils import BaseTestCase


@pytest.mark.client_mark
class TestStatuses(BaseTestCase):
    @mock.patch("polyaxon_sdk.RunsV1Api.get_run_statuses")
    def test_get_statuses(self, sdk_get_run_statuses):
        client = RunClient(owner="owner", project="project", run_uuid=uuid.uuid4().hex)
        for _ in client.get_statuses():
            pass
        assert sdk_get_run_statuses.call_count == 1

    @mock.patch("polyaxon_sdk.RunsV1Api.get_run_statuses")
    def test_get_statuses_watch(self, sdk_get_run_statuses):
        settings.CLIENT_CONFIG.watch_interval = 1
        client = RunClient(owner="owner", project="project", run_uuid=uuid.uuid4().hex)
        for _ in client.watch_statuses():
            resp = MagicMock(status=V1Statuses.FAILED, status_conditions=[])
            sdk_get_run_statuses.return_value = resp
        assert sdk_get_run_statuses.call_count == 2



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from polyaxon.logger import logger


@click.group()
def tuner():
    pass


@tuner.command()
@click.option(
    "--matrix",
    help="A string representing the matrix configuration for bayesian optimization.",
)
@click.option(
    "--join", help="A string representing the join to fetch configs and metrics."
)
@click.option("--iteration", type=int, help="The current iteration.")
def bayes(matrix, join, iteration):
    """Create suggestions based on bayesian optimization."""
    from polyaxon.client import RunClient
    from polyaxon.polyflow import V1Bayes, V1Join
    from polyaxon.polytune.iteration_lineage import (
        get_iteration_definition,
        handle_iteration,
        handle_iteration_failure,
    )
    from polyaxon.polytune.search_managers.bayesian_optimization.manager import (
        BayesSearchManager,
    )

    matrix = V1Bayes.read(matrix)
    join = V1Join.read(join)
    client = RunClient()
    values = get_iteration_definition(
        client=client,
        iteration=iteration,
        join=join,
        optimization_metric=matrix.metric.name,
    )
    if not values:
        return
    run_uuids, configs, metrics = values

    retry = 1
    exp = None
    suggestions = None
    while retry < 3:
        try:
            suggestions = BayesSearchManager(
                config=matrix,
            ).get_suggestions(configs=configs, metrics=metrics)
            exp = None
            break
        except Exception as exp:
            retry += 1
            logger.warning(exp)

    if exp:
        handle_iteration_failure(client=client, exp=exp)
        return

    handle_iteration(
        client=client,
        iteration=iteration,
        suggestions=suggestions,
    )


@tuner.command()
@click.option(
    "--matrix", help="A string representing the matrix configuration for hyperband."
)
@click.option(
    "--join", help="A string representing the join to fetch configs and metrics."
)
@click.option("--iteration", type=int, help="The current hyperband iteration.")
@click.option(
    "--bracket-iteration", type=int, help="The current hyperband bracket iteration."
)
def hyperband(matrix, join, iteration, bracket_iteration):
    """Create suggestions based on hyperband."""
    from polyaxon.client import RunClient
    from polyaxon.polyflow import V1Hyperband, V1Join
    from polyaxon.polytune.iteration_lineage import (
        get_iteration_definition,
        handle_iteration,
        handle_iteration_failure,
    )
    from polyaxon.polytune.search_managers.hyperband.manager import HyperbandManager

    matrix = V1Hyperband.read(matrix)
    matrix.set_tuning_params()
    join = V1Join.read(join)
    client = RunClient()
    values = get_iteration_definition(
        client=client,
        iteration=iteration,
        join=join,
        optimization_metric=matrix.metric.name,
        name="in-iteration-{}-{}".format(iteration, bracket_iteration),
    )
    if not values:
        return
    run_uuids, configs, metrics = values

    retry = 1
    exp = None
    suggestions = None
    while retry < 3:
        try:
            suggestions = HyperbandManager(config=matrix).get_suggestions(
                configs=configs,
                metrics=metrics,
                bracket_iteration=bracket_iteration,
                iteration=iteration,
            )
            exp = None
            break
        except Exception as exp:
            retry += 1
            logger.warning(exp)

    if exp:
        handle_iteration_failure(client=client, exp=exp)
        return

    handle_iteration(
        client=client,
        iteration=iteration,
        suggestions=suggestions,
        summary={"bracket_iteration": bracket_iteration},
        name="out-iteration-{}-{}".format(iteration, bracket_iteration),
    )


@tuner.command()
@click.option(
    "--matrix", help="A string representing the matrix configuration for hyperopt."
)
@click.option(
    "--join", help="A string representing the join to fetch configs and metrics."
)
@click.option("--iteration", type=int, help="The current iteration.")
def hyperopt(matrix, join, iteration):
    """Create suggestions based on hyperopt."""
    from polyaxon.client import RunClient
    from polyaxon.polyflow import V1Hyperopt, V1Join
    from polyaxon.polytune.iteration_lineage import (
        get_iteration_definition,
        handle_iteration,
        handle_iteration_failure,
    )
    from polyaxon.polytune.search_managers.hyperopt.manager import HyperoptManager

    matrix = V1Hyperopt.read(matrix)
    join = V1Join.read(join)
    client = RunClient()
    values = get_iteration_definition(
        client=client,
        iteration=iteration,
        join=join,
        optimization_metric=matrix.metric.name,
    )
    if not values:
        return
    run_uuids, configs, metrics = values

    retry = 1
    exp = None
    suggestions = None
    while retry < 3:
        try:
            suggestions = HyperoptManager(config=matrix).get_suggestions(
                configs=configs, metrics=metrics
            )
            exp = None
            break
        except Exception as exp:
            retry += 1
            logger.warning(exp)

    if exp:
        handle_iteration_failure(client=client, exp=exp)
        return

    handle_iteration(
        client=client,
        iteration=iteration,
        suggestions=suggestions,
    )



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools

from polyaxon_sdk.rest import ApiException
from urllib3.exceptions import HTTPError

from polyaxon import settings
from polyaxon.logger import logger


def client_handler(
    check_no_op: bool = True,
    check_offline: bool = False,
    can_log_events: bool = False,
    can_log_outputs: bool = False,
):
    def client_handler_wrapper(f):
        """
        The `ClientHandlerDecorator` is a decorator to handle several checks in PolyaxonClient.

         * check_offline: to ignore any decorated function when POLYAXON_IS_OFFLINE env var is found.
         * check_no_op: to ignore any decorated function when NO_OP env var is found.
         * handle_openapi_exceptions: to handle exception of OpenApi and generate better
            debugging outputs.
         * can_log_events: to check if there's an event logger instance on the object.
         * can_log_outputs: to check if there's an outputs path set on the run.
         * openapi_extra_context: to augment openapi errors.

        usage example with class method:
            @client_handler
            def my_func(self, *args, **kwargs):
                ...
                return ...

        usage example with a function:
            @client_handler
            def my_func(arg1, arg2):
                ...
                return ...
        """

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if check_no_op and settings.CLIENT_CONFIG.no_op:
                logger.debug("Using NO_OP mode")
                return None
            if check_offline:
                self_arg = args[0] if args else None
                if settings.CLIENT_CONFIG.is_offline or getattr(
                    self_arg, "_is_offline", False
                ):
                    logger.debug("Using IS_OFFLINE mode")
                    return None
            if args:
                self_arg = args[0]

                if can_log_events and (
                    not hasattr(self_arg, "_event_logger")
                    or self_arg._event_logger is None  # pylint:disable=protected-access
                ):
                    logger.warning(
                        "You should set an event logger before calling: {}".format(
                            repr(f)
                        )
                    )

                if can_log_outputs and (
                    not hasattr(self_arg, "_outputs_path")
                    or self_arg._outputs_path is None  # pylint:disable=protected-access
                ):
                    logger.warning(
                        "You should set an an outputs path before calling: {}".format(
                            repr(f)
                        )
                    )

            try:
                return f(*args, **kwargs)
            except (ApiException, HTTPError) as e:
                logger.debug("Client config:\n%s\n", settings.CLIENT_CONFIG.to_dict())
                raise e

        return wrapper

    return client_handler_wrapper



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.exceptions import PolyaxonSchemaError


def get_queue_info(queue: str):
    if not queue:
        raise PolyaxonSchemaError("Received an invalid queue {}".format(queue))

    parts = queue.replace(".", "/").split("/")
    agent = None
    queue_name = queue
    if len(parts) == 2:
        agent, queue_name = parts
    elif len(parts) > 2:
        raise PolyaxonSchemaError(
            "Please provide a valid queue. "
            "The queue name should be: queue-name to use the default agent or agent-name/queue."
        )

    return agent, queue_name



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.env_vars.getters import get_component_info, get_versioned_entity_full_name
from polyaxon.exceptions import PolyaxonSchemaError
from tests.utils import BaseTestCase


class TestComponentEnvVars(BaseTestCase):
    def test_get_versioned_entity_full_name(self):
        assert get_versioned_entity_full_name(None) is None
        assert get_versioned_entity_full_name("hub") == "hub"
        assert get_versioned_entity_full_name("hub", tag="ver") == "hub:ver"
        assert get_versioned_entity_full_name("hub", "owner") == "owner/hub"
        assert get_versioned_entity_full_name("hub", "owner", "ver") == "owner/hub:ver"

    def test_get_component_info(self):
        with self.assertRaises(PolyaxonSchemaError):
            get_component_info(None)

        with self.assertRaises(PolyaxonSchemaError):
            get_component_info("")

        assert get_component_info("hub") == ("polyaxon", "hub", "latest")
        assert get_component_info("hub:ver") == ("polyaxon", "hub", "ver")
        assert get_component_info("owner.hub") == ("owner", "hub", "latest")
        assert get_component_info("owner/hub") == ("owner", "hub", "latest")
        assert get_component_info("owner.hub:ver") == ("owner", "hub", "ver")
        assert get_component_info("owner/hub:ver") == ("owner", "hub", "ver")



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import uuid

from polyaxon.env_vars.keys import POLYAXON_KEYS_RUN_INSTANCE
from polyaxon.exceptions import PolyaxonContainerException
from polyaxon.sidecar import start_sidecar
from tests.utils import patch_settings


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
async def test_monitor_raise_if_no_env_is_set():
    patch_settings()
    os.environ[POLYAXON_KEYS_RUN_INSTANCE] = "foo"
    with pytest.raises(PolyaxonContainerException):
        await start_sidecar(
            container_id="foo",
            sleep_interval=3,
            sync_interval=6,
            monitor_outputs=True,
            monitor_logs=False,
        )
    del os.environ[POLYAXON_KEYS_RUN_INSTANCE]


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
async def test_monitor_raise_if_no_pod_id():
    patch_settings()
    os.environ[POLYAXON_KEYS_RUN_INSTANCE] = "owner.project.runs.{}".format(
        uuid.uuid4().hex
    )
    with pytest.raises(PolyaxonContainerException):
        await start_sidecar(
            container_id="foo",
            sleep_interval=3,
            sync_interval=6,
            monitor_outputs=True,
            monitor_logs=False,
        )
    del os.environ[POLYAXON_KEYS_RUN_INSTANCE]



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import ujson

from starlette import status
from starlette.datastructures import UploadFile
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response

from polyaxon import settings
from polyaxon.stores.async_manager import upload_dir, upload_file
from polyaxon.utils.path_utils import check_or_create_path, delete_path, untar_file


async def handle_posted_data(
    content_file: UploadFile,
    root_path: str,
    path: str,
    upload: bool,
    is_file: bool,
    overwrite: bool = True,
    untar: bool = True,
) -> str:
    tmp_path = "{}/{}".format(
        root_path, os.path.basename(content_file.filename)
    ).rstrip("/")
    if path:
        root_path = "{}/{}".format(root_path, path).rstrip("/")
        if is_file:
            root_path = "{}/{}".format(
                root_path, os.path.basename(content_file.filename)
            )
    else:
        if untar:
            root_path = "{}/{}".format(root_path, "uploads").rstrip("/")
        else:
            root_path = tmp_path
    if not untar:
        tmp_path = root_path
    full_tmppath = os.path.join(settings.AGENT_CONFIG.artifacts_root, tmp_path)
    full_filepath = os.path.join(settings.AGENT_CONFIG.artifacts_root, root_path)

    if overwrite and os.path.exists(full_filepath):
        delete_path(full_filepath)
    if not overwrite and os.path.exists(full_filepath):
        return full_filepath
    # Always clean tmp path
    if overwrite and os.path.exists(full_tmppath):
        delete_path(full_tmppath)

    check_or_create_path(full_tmppath, is_dir=False)
    check_or_create_path(full_filepath, is_dir=not is_file)

    # Creating the new file
    with open(full_tmppath, "wb") as destination:
        for chunk in content_file.file:
            destination.write(chunk)
    if untar:
        untar_file(full_tmppath, extract_path=full_filepath, use_filepath=False)
    if upload:
        if is_file:
            await upload_file(subpath=root_path)
        else:
            await upload_dir(subpath=root_path)
    return root_path


async def handle_upload(request: Request, is_file: bool) -> Response:
    form = await request.form()
    content_file = form["upload_file"]  # type: UploadFile
    content_json = form["json"]  # type: str
    content_json = ujson.loads(content_json) if content_json else {}
    run_uuid = request.path_params["run_uuid"]
    overwrite = content_json.get("overwrite", True)
    untar = content_json.get("untar", True)
    path = content_json.get("path", "")
    try:
        archived_path = await handle_posted_data(
            content_file=content_file,
            root_path=run_uuid,
            path=path,
            upload=True,
            is_file=is_file,
            overwrite=overwrite,
            untar=untar,
        )
    except Exception as e:
        raise HTTPException(
            detail="Run's artifacts upload was unsuccessful, "
            "an error was raised while uploading the data %s." % e,
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    if not archived_path:
        return Response(
            content="Artifact not found and not uploaded: filepath={}".format(
                archived_path
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return Response(status_code=status.HTTP_200_OK)



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import click

from marshmallow import ValidationError

from polyaxon import settings
from polyaxon.cli.executor import docker_run, k8s_run, platform_run
from polyaxon.cli.options import OPTIONS_PROJECT
from polyaxon.env_vars.getters import get_project_or_local
from polyaxon.exceptions import PolyaxonSchemaError
from polyaxon.logger import clean_outputs
from polyaxon.managers.git import GitConfigManager
from polyaxon.polyaxonfile import (
    CompiledOperationSpecification,
    OperationSpecification,
    check_polyaxonfile,
)
from polyaxon.utils import code_reference
from polyaxon.utils.formatting import Printer
from polyaxon.utils.validation import validate_tags


@click.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(
    "-f",
    "--file",
    "polyaxonfile",
    multiple=True,
    type=click.Path(exists=True),
    help="The polyaxonfiles to run.",
)
@click.option(
    "-pm",
    "--python-module",
    type=str,
    help="The python module containing the polyaxonfile to run.",
)
@click.option(
    "--url",
    type=str,
    help="The url containing the polyaxonfile to run.",
)
@click.option(
    "--hub",
    type=str,
    help="The Component Hub name containing the polyaxonfile to run.",
)
@click.option(
    "--name",
    type=str,
    help="Name to give to this run, must be unique within the project, could be none.",
)
@click.option("--tags", type=str, help="Tags of this run, comma separated values.")
@click.option("--description", type=str, help="The description to give to this run.")
@click.option(
    "--log",
    "-l",
    is_flag=True,
    default=False,
    help="To start logging after scheduling the run.",
)
@click.option(
    "--upload",
    "-u",
    is_flag=True,
    default=False,
    help="To upload an init context before scheduling the run.",
)
@click.option(
    "--upload-from",
    "-u-from",
    type=str,
    help="The path to upload from relative the current location.",
)
@click.option(
    "--upload-to",
    "-u-to",
    type=str,
    help="The path to upload to relative the run's root context.",
)
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    default=False,
    help="To start statuses watch loop after scheduling the run.",
)
@click.option(
    "--local",
    is_flag=True,
    default=False,
    help="To start the run locally, with `docker` environment as default.",
)
@click.option(
    "--params",
    "-P",
    metavar="NAME=VALUE",
    multiple=True,
    help="A parameter to override the default params of the run, form -P name=value.",
)
@click.option("--presets", type=str, help="Name of the presets to use for this run.")
@click.option(
    "--queue",
    "-q",
    type=str,
    help="Name of the queue to use for this run. "
    "If the name is not namespaced by the agent name the default agent is used: "
    "queue-name or agent-name/queue-name",
)
@click.option(
    "--nocache",
    is_flag=True,
    default=False,
    help="Disable cache check before starting this operation.",
)
@click.option(
    "--cache",
    is_flag=True,
    default=False,
    help="Enable cache check before starting this operation.",
)
@click.option(
    "--eager",
    is_flag=True,
    default=False,
    help="A flag to enable eager mode for pipeline operations, "
    "currently this mode supports grid search, random search, and parallel mapping. "
    "Note that this flag requires numpy.",
)
@click.option(
    "--git-preset",
    is_flag=True,
    default=False,
    help="A flag to enable automatic injection of a git initializer as a preset "
    "using the initialized git connection.",
)
@click.option(
    "--git-revision",
    type=str,
    help="If provided, Polyaxon will use this git revision "
    "instead of trying to detected and use the latest commit. "
    "The git revision could be a commit or a branch or any valid tree-ish.",
)
@click.option(
    "--ignore-template",
    is_flag=True,
    default=False,
    help="If provided, Polyaxon will ignore template definition and "
    "submit the manifest to be checked by the API.",
)
@click.pass_context
@clean_outputs
def run(
    ctx,
    project,
    polyaxonfile,
    python_module,
    url,
    hub,
    name,
    tags,
    description,
    log,
    upload,
    upload_from,
    upload_to,
    watch,
    local,
    params,
    presets,
    queue,
    nocache,
    cache,
    eager,
    git_preset,
    git_revision,
    ignore_template,
):
    """Run polyaxonfile specification.

    Examples:

    \b
    $ polyaxon run -f file -f file_override ...

    Run and set description and tags for this run

    \b
    $ polyaxon run -f file --description="Description of the current run" --tags="foo, bar, moo"

    Run and set a unique name for this run

    \b
    polyaxon run --name=foo

    Run for a specific project

    \b
    $ polyaxon run -p project1 -f file.yaml

    Run with updated params

    \b
    $ polyaxon run -p project1 -f file.yaml -P param1=234.2 -P param2=relu

    If a python file contains a component main, you can run that component

    \b
    polyaxon run -pm path/to/my-component.py


    If a python file contains more than one component, you can specify the component to run

    \b
    polyaxon run -pm path/to/my-component.py:componentA
    """
    if cache and nocache:
        Printer.print_error(
            "You can't use `--cache` and `--nocache` at the same.", sys_exit=True
        )
    if (upload_to or upload_from) and not upload:
        upload = True
    if upload and eager:
        Printer.print_error(
            "You can't use `--upload` and `--eager` at the same.", sys_exit=True
        )

    git_init = None
    if git_preset:
        # Check that the current path was initialized
        if not GitConfigManager.is_initialized():
            Printer.print_error(
                "You can't use --git-init, "
                "the current path is not initialized with a valid git connection or a git url, "
                "please run `polyaxon init [--git-connection] [--git-url]` "
                "to set a valid git configuration.",
                sys_exit=True,
            )
        git_init = GitConfigManager.get_config()
        if git_revision:
            git_init.git.revision = git_revision
        elif code_reference.is_git_initialized(path="."):
            if code_reference.is_dirty(path="."):
                Printer.print_warning(
                    "Polyaxon detected uncommitted changes in the current git repo!"
                )
            commit_hash = code_reference.get_commit()
            git_init.git.revision = commit_hash
        else:
            Printer.print_warning(
                "Polyaxon could not find a valid git repo, "
                "and will not add the current commit to the git initializer."
            )

    presets = validate_tags(presets)

    op_spec = check_polyaxonfile(
        polyaxonfile=polyaxonfile,
        python_module=python_module,
        url=url,
        hub=hub,
        params=params,
        presets=presets,
        queue=queue,
        cache=cache,
        nocache=nocache,
        verbose=False,
        eager=eager,
        git_init=git_init,
        ignore_template=ignore_template,
    )

    if ignore_template:
        op_spec.disable_template()
    if op_spec.is_template():
        click.echo("Please customize the specification or disable the template.")
        sys.exit(1)

    owner, project_name = get_project_or_local(project, is_cli=True)
    tags = validate_tags(tags)

    if local:
        try:
            compiled_operation = OperationSpecification.compile_operation(op_spec)
            compiled_operation = (
                CompiledOperationSpecification.apply_operation_contexts(
                    compiled_operation
                )
            )
        except (PolyaxonSchemaError, ValidationError):
            Printer.print_error(
                "Could not run this polyaxonfile locally, "
                "a context is required to resolve it dependencies."
            )
            sys.exit(1)
        docker_run(
            ctx=ctx,
            name=name,
            owner=owner,
            project_name=project_name,
            description=description,
            tags=tags,
            compiled_operation=compiled_operation,
            log=log,
        )
    elif settings.CLIENT_CONFIG.no_api:
        k8s_run(
            ctx=ctx,
            name=name,
            owner=owner,
            project_name=project_name,
            description=description,
            tags=tags,
            op_spec=op_spec,
            log=log,
        )
    else:
        platform_run(
            ctx=ctx,
            name=name,
            owner=owner,
            project_name=project_name,
            description=description,
            tags=tags,
            op_spec=op_spec,
            log=log,
            upload=upload,
            upload_to=upload_to,
            upload_from=upload_from,
            watch=watch,
            eager=eager,
        )



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from typing import Dict

from polyaxon import settings
from polyaxon.client import RunClient
from polyaxon.exceptions import PolyaxonClientException, PolyaxonContainerException
from polyaxon.polyboard.artifacts import V1ArtifactKind, V1RunArtifact


def create_file_lineage(filepath: str, summary: Dict, kind: str):
    kind = kind or V1ArtifactKind.FILE

    if not filepath:
        return
    filename = os.path.basename(filepath)

    if settings.CLIENT_CONFIG.no_api:
        return

    try:
        run_client = RunClient()
    except PolyaxonClientException as e:
        raise PolyaxonContainerException(e)

    artifact_run = V1RunArtifact(
        name=filename,
        kind=kind,
        path=RunClient.get_rel_asset_path(filepath),
        summary=summary,
        is_input=True,
    )
    run_client.log_artifact_lineage(artifact_run)



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from unittest.mock import patch

from polyaxon.polyflow.matrix import V1RandomSearch
from polyaxon.polytune.search_managers.random_search.manager import RandomSearchManager
from tests.utils import BaseTestCase


@pytest.mark.polytune_mark
class TestRandomSearch(BaseTestCase):
    def test_random_search_config(self):
        assert RandomSearchManager.CONFIG == V1RandomSearch

    def test_get_suggestions(self):
        config = V1RandomSearch.from_dict(
            {
                "concurrency": 2,
                "numRuns": 10,
                "params": {
                    "feature1": {"kind": "choice", "value": [1, 2]},
                    "feature3": {"kind": "range", "value": [1, 3, 1]},
                },
            }
        )

        assert len(RandomSearchManager(config).get_suggestions()) == 4

        config = V1RandomSearch.from_dict(
            {
                "concurrency": 2,
                "numRuns": 10,
                "params": {
                    "feature1": {"kind": "pchoice", "value": [(1, 0.1), (2, 0.6)]},
                    "feature3": {"kind": "range", "value": [1, 3, 1]},
                },
            }
        )

        assert len(RandomSearchManager(config).get_suggestions()) == 4

        config = V1RandomSearch.from_dict(
            {
                "concurrency": 2,
                "numRuns": 10,
                "params": {
                    "feature1": {"kind": "choice", "value": [1, 2, 3]},
                    "feature2": {"kind": "linspace", "value": [1, 2, 5]},
                    "feature3": {"kind": "range", "value": [1, 5, 1]},
                },
            }
        )
        assert len(RandomSearchManager(config).get_suggestions()) == 10

        config = V1RandomSearch.from_dict(
            {
                "concurrency": 2,
                "numRuns": 10,
                "params": {
                    "feature1": {
                        "kind": "pchoice",
                        "value": [(1, 0.3), (2, 0.3), (3, 0.3)],
                    },
                    "feature2": {"kind": "uniform", "value": [0, 1]},
                    "feature3": {"kind": "qlognormal", "value": [0, 0.5, 0.51]},
                },
            }
        )
        assert len(RandomSearchManager(config).get_suggestions()) == 10

    def test_get_suggestions_calls_sample(self):
        config = V1RandomSearch.from_dict(
            {
                "concurrency": 2,
                "numRuns": 1,
                "params": {
                    "feature1": {"kind": "choice", "value": [1, 2, 3]},
                    "feature2": {"kind": "linspace", "value": [1, 2, 5]},
                    "feature3": {"kind": "range", "value": [1, 5, 1]},
                },
            }
        )
        with patch(
            "polyaxon.polytune.search_managers.random_search.manager.sample"
        ) as sample_mock:
            RandomSearchManager(config).get_suggestions()

        assert sample_mock.call_count == 3

        config = V1RandomSearch.from_dict(
            {
                "concurrency": 2,
                "numRuns": 1,
                "params": {
                    "feature1": {
                        "kind": "pchoice",
                        "value": [(1, 0.3), (2, 0.3), (3, 0.3)],
                    },
                    "feature2": {"kind": "uniform", "value": [0, 1]},
                    "feature3": {"kind": "qlognormal", "value": [0, 0.5, 0.51]},
                    "feature4": {"kind": "range", "value": [1, 5, 1]},
                },
            }
        )
        with patch(
            "polyaxon.polytune.search_managers.random_search.manager.sample"
        ) as sample_mock:
            RandomSearchManager(config).get_suggestions()

        assert sample_mock.call_count == 4



#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys

import click

from polyaxon_sdk.rest import ApiException
from urllib3.exceptions import HTTPError

from polyaxon.api import REWRITE_SERVICES_V1, SERVICES_V1
from polyaxon.cli.dashboard import get_dashboard, get_dashboard_url
from polyaxon.cli.errors import handle_cli_error
from polyaxon.cli.options import OPTIONS_PROJECT, OPTIONS_RUN_UID
from polyaxon.client import RunClient
from polyaxon.client.run import get_run_logs
from polyaxon.containers import contexts as container_contexts
from polyaxon.env_vars.getters import get_project_or_local, get_project_run_or_local
from polyaxon.exceptions import (
    PolyaxonClientException,
    PolyaxonHTTPError,
    PolyaxonShouldExitError,
)
from polyaxon.lifecycle import LifeCycle, V1Statuses
from polyaxon.logger import clean_outputs
from polyaxon.managers.run import RunConfigManager
from polyaxon.metadata.keys import META_REWRITE_PATH
from polyaxon.polyaxonfile import OperationSpecification
from polyaxon.polyflow import V1RunKind
from polyaxon.utils import cache
from polyaxon.utils.csv_utils import write_csv
from polyaxon.utils.formatting import (
    Printer,
    dict_tabulate,
    dict_to_tabulate,
    flatten_keys,
    get_meta_response,
    list_dicts_to_csv,
    list_dicts_to_tabulate,
)
from polyaxon.utils.validation import validate_tags

DEFAULT_EXCLUDE = [
    "owner",
    "project",
    "description",
    "content",
    "raw_content",
    "live_state",
    "readme",
    "settings",
    "meta_info",
    "is_approved",
    "is_managed",
    "schedule_at",
    "original",
    "pipeline",
    "role",
    "status_conditions",
]


def get_run_details(run):  # pylint:disable=redefined-outer-name
    if run.description:
        Printer.print_header("Run description:")
        click.echo("{}\n".format(run.description))

    if run.inputs:
        Printer.print_header("Run inputs:")
        dict_tabulate(run.inputs)

    if run.outputs:
        Printer.print_header("Run outputs:")
        dict_tabulate(run.outputs)

    if run.settings:
        Printer.print_header("Run settings:")
        dict_tabulate(run.settings.to_dict())

    if run.meta_info:
        Printer.print_header("Run meta info:")
        dict_tabulate(run.meta_info)

    response = Printer.add_status_color(run.to_dict())
    response = dict_to_tabulate(
        response,
        humanize_values=True,
        exclude_attrs=[
            "project",
            "description",
            "readme",
            "content",
            "raw_content",
            "inputs",
            "outputs",
            "is_managed",
            "status_conditions",
            "settings",
            "meta_info",
        ],
    )

    Printer.print_header("Run info:")
    dict_tabulate(response)


@click.group()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.pass_context
@clean_outputs
def ops(ctx, project, uid):
    """Commands for ops/runs."""
    ctx.obj = ctx.obj or {}
    if project or uid:
        Printer.print_warning(
            "Passing arguments to command groups is deprecated and will be removed in v1.2! "
            "Please use arguments on the sub-command directly: `polyaxon ops SUB_COMMAND --help`"
        )
    ctx.obj["project"] = project
    if ctx.invoked_subcommand not in ["ls"]:
        ctx.obj["run_uuid"] = uid


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(
    "--io",
    "-io",
    is_flag=True,
    help="List runs with their inputs/outputs (params, metrics, results, ...).",
)
@click.option(
    "--to-csv",
    is_flag=True,
    help="Saves the results to a csv file. Note that this flag requires pandas",
)
@click.option(
    "--query", "-q", type=str, help="To filter the runs based on this query spec."
)
@click.option(
    "--sort", "-s", type=str, help="To order the runs based on this sort spec."
)
@click.option("--limit", "-l", type=int, help="To limit the list of runs.")
@click.option("--offset", "-off", type=int, help="To offset the list of runs.")
@click.option("--columns", "-c", type=str, help="The columns to show.")
@click.option(
    "--offline",
    is_flag=True,
    default=False,
    help="To list offline runs if they exists.",
)
@click.option(
    "--offline-path",
    type=click.Path(exists=False),
    help="Optional path to use to where offline runs are persisted, "
    "default value is taken from the env var: `POLYAXON_OFFLINE_ROOT`.",
)
@click.pass_context
@clean_outputs
def ls(
    ctx, project, io, to_csv, query, sort, limit, offset, columns, offline, offline_path
):
    """List runs for this project.

    Uses /docs/core/cli/#caching

    Examples:

    Get all runs:

    \b

    Get all runs with with status {created or running}, and
    creation date between 2018-01-01 and 2018-01-02, and params activation equal to sigmoid
    and metric loss less or equal to 0.2

    \b
    $ polyaxon ops ls \
    -q "status:created|running, started_at:2018-01-01..2018-01-02, \
    params.activation:sigmoid, metrics.loss:<=0.2"


    Get all runs sorted by update date:

    \b
    $ polyaxon ops ls -s "-updated_at"

    Get all runs of kind job:

    \b
    $ polyaxon ops ls -q "kind: job"

    Get all runs of kind service:

    \b
    $ polyaxon ops ls -q "kind: service"
    """
    if offline:
        offline_path = offline_path or container_contexts.CONTEXT_OFFLINE_ROOT
        offline_path_format = "{}/{{}}/run_data.json".format(offline_path)
        if not os.path.exists(offline_path) or not os.path.isdir(offline_path):
            Printer.print_error(
                f"Could not list offline runs, the path `{offline_path}` "
                f"does not exist or is not a directory."
            )
            sys.exit(1)
        results = []
        for uid in os.listdir(offline_path):
            run_path = offline_path_format.format(uid)
            if os.path.exists(run_path):
                results.append(RunConfigManager.read_from_path(run_path))
            else:
                Printer.print_warning(f"Skipping run {uid}, offline data not found.")
    else:
        owner, project_name = get_project_or_local(
            project or ctx.obj.get("project"), is_cli=True
        )

        try:
            polyaxon_client = RunClient(owner=owner, project=project_name)
            response = polyaxon_client.list(
                limit=limit, offset=offset, query=query, sort=sort
            )
        except (ApiException, HTTPError) as e:
            handle_cli_error(
                e, message="Could not get runs for project `{}`.".format(project_name)
            )
            sys.exit(1)

        meta = get_meta_response(response)
        if meta:
            Printer.print_header(
                "Runs for project `{}/{}`.".format(owner, project_name)
            )
            Printer.print_header("Navigation:")
            dict_tabulate(meta)
        else:
            Printer.print_header(
                "No runs found for project `{}/{}`.".format(owner, project_name)
            )

        results = response.results

    objects = [Printer.add_status_color(o.to_dict()) for o in results]
    columns = validate_tags(columns)
    if io:
        objects, prefixed_columns = flatten_keys(
            objects=objects,
            columns=["inputs", "outputs"],
            columns_prefix={"inputs": "in", "outputs": "out"},
        )
        if columns:
            columns = {prefixed_columns.get(col, col) for col in columns}
        if to_csv:
            objects = list_dicts_to_csv(
                objects,
                include_attrs=columns,
                exclude_attrs=DEFAULT_EXCLUDE,
            )
        else:
            objects = list_dicts_to_tabulate(
                objects,
                include_attrs=columns,
                exclude_attrs=DEFAULT_EXCLUDE,
                humanize_values=True,
                upper_keys=True,
            )
    else:
        if to_csv:
            objects = list_dicts_to_csv(
                objects,
                include_attrs=columns,
                exclude_attrs=DEFAULT_EXCLUDE + ["inputs", "outputs"],
            )
        else:
            objects = list_dicts_to_tabulate(
                objects,
                include_attrs=columns,
                exclude_attrs=DEFAULT_EXCLUDE + ["inputs", "outputs"],
                humanize_values=True,
                upper_keys=True,
            )
    if objects:
        if to_csv:
            filename = "./results.csv"
            write_csv(objects, filename=filename)
            Printer.print_success("CSV file generated: `{}`".format(filename))
        else:
            Printer.print_header("Runs:")
            objects.pop("project_name", None)
            dict_tabulate(objects, is_list_dict=True)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--offline",
    is_flag=True,
    default=False,
    help="To list offline runs if they exists.",
)
@click.option(
    "--offline-path",
    type=click.Path(exists=False),
    help="Optional path to use to where offline runs are persisted, "
    "default value is taken from the env var: `POLYAXON_OFFLINE_ROOT`.",
)
@click.pass_context
@clean_outputs
def get(ctx, project, uid, offline, offline_path):
    """Get run.

    Uses /docs/core/cli/#caching

    Examples for getting a run:

    \b
    $ polyaxon ops get  # if run is cached

    \b
    $ polyaxon ops get --uid=8aac02e3a62a4f0aaa257c59da5eab80 # project is cached

    \b
    $ polyaxon ops get --project=cats-vs-dogs -uid 8aac02e3a62a4f0aaa257c59da5eab80

    \b
    $ polyaxon ops get -p alain/cats-vs-dogs --uid=8aac02e3a62a4f0aaa257c59da5eab80
    """

    uid = uid or ctx.obj.get("run_uuid")

    if offline:
        offline_path = offline_path or container_contexts.CONTEXT_OFFLINE_ROOT
        offline_path = "{}/{}/run_data.json".format(offline_path, uid)
        if not os.path.exists(offline_path):
            Printer.print_error(
                f"Could not get offline run, the path `{offline_path}` "
                f"does not exist."
            )
            sys.exit(1)
        run_data = RunConfigManager.read_from_path(offline_path)
    else:
        owner, project_name, run_uuid = get_project_run_or_local(
            project or ctx.obj.get("project"),
            uid,
            is_cli=True,
        )

        try:
            polyaxon_client = RunClient(
                owner=owner, project=project_name, run_uuid=run_uuid
            )
            polyaxon_client.refresh_data()
            config = polyaxon_client.client.sanitize_for_serialization(
                polyaxon_client.run_data
            )
            cache.cache(
                config_manager=RunConfigManager,
                config=config,
                owner=owner,
                project=project_name,
            )
            run_data = polyaxon_client.run_data
        except (ApiException, HTTPError) as e:
            handle_cli_error(
                e,
                message="Could not load run `{}/{}/{}` info.".format(
                    owner, project_name, run_uuid
                ),
            )
            sys.exit(1)

    get_run_details(run_data)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.pass_context
@clean_outputs
def delete(ctx, project, uid):
    """Delete a run.

    Uses /docs/core/cli/#caching

    Example:

    \b
    $ polyaxon ops delete

    \b
    $ polyaxon ops delete --uid=8aac02e3a62a4f0aaa257c59da5eab80  # project is cached

    \b
    $ polyaxon ops delete --project=cats-vs-dogs -uid 8aac02e3a62a4f0aaa257c59da5eab80
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    if not click.confirm("Are sure you want to delete run `{}`".format(run_uuid)):
        click.echo("Existing without deleting the run.")
        sys.exit(1)

    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        polyaxon_client.delete()
        # Purge caching
        RunConfigManager.purge()
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not delete run `{}`.".format(run_uuid))
        sys.exit(1)

    Printer.print_success("Run `{}` was delete successfully".format(run_uuid))


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option("--name", type=str, help="Name of the run (optional).")
@click.option("--description", type=str, help="Description of the run (optional).")
@click.option(
    "--tags", type=str, help="Tags of the run, comma separated values (optional)."
)
@click.pass_context
@clean_outputs
def update(ctx, project, uid, name, description, tags):
    """Update run.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops  update --uid=8aac02e3a62a4f0aaa257c59da5eab80
    --description="new description for my runs"

    \b
    $ polyaxon ops  update --project=cats-vs-dogs -uid 8aac02e3a62a4f0aaa257c59da5eab80
    --tags="foo, bar" --name="unique-name"
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    update_dict = {}

    if name:
        update_dict["name"] = name

    if description:
        update_dict["description"] = description

    tags = validate_tags(tags)
    if tags:
        update_dict["tags"] = tags

    if not update_dict:
        Printer.print_warning("No argument was provided to update the run.")
        sys.exit(0)

    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        response = polyaxon_client.update(update_dict)
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not update run `{}`.".format(run_uuid))
        sys.exit(1)

    Printer.print_success("Run updated.")
    get_run_details(response)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.pass_context
@clean_outputs
def approve(ctx, project, uid):
    """Approve a run waiting for human validation.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops approve

    \b
    $ polyaxon ops approve --uid=8aac02e3a62a4f0aaa257c59da5eab80
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )

    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        polyaxon_client.approve()
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not approve run `{}`.".format(run_uuid))
        sys.exit(1)

    Printer.print_success("Run is approved.")


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Automatic yes to prompts. "
    'Assume "yes" as answer to all prompts and run non-interactively.',
)
@click.pass_context
@clean_outputs
def stop(ctx, project, uid, yes):
    """Stop run.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops stop

    \b
    $ polyaxon ops stop --uid=8aac02e3a62a4f0aaa257c59da5eab80
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    if not yes and not click.confirm(
        "Are sure you want to stop " "run `{}`".format(run_uuid)
    ):
        click.echo("Existing without stopping run.")
        sys.exit(0)

    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        polyaxon_client.stop()
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not stop run `{}`.".format(run_uuid))
        sys.exit(1)

    Printer.print_success("Run is being stopped.")


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--copy",
    "-c",
    is_flag=True,
    default=False,
    help="To copy the run before restarting.",
)
@click.option(
    "-f",
    "--file",
    "polyaxonfile",
    multiple=True,
    type=click.Path(exists=True),
    help="The polyaxonfiles to update with, they should be an operation preset.",
)
@click.pass_context
@clean_outputs
def restart(ctx, project, uid, copy, polyaxonfile):
    """Restart run.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon run --uid=8aac02e3a62a4f0aaa257c59da5eab80 restart
    """
    content = None
    if polyaxonfile:
        content = OperationSpecification.read(polyaxonfile, is_preset=True).to_dict(
            dump=True
        )

    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        response = polyaxon_client.restart(override_config=content, copy=copy)
        Printer.print_success(
            "Run was {} with uid {}".format(
                "copied" if copy else "restarted", response.uuid
            )
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not restart run `{}`.".format(run_uuid))
        sys.exit(1)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "-f",
    "--file",
    "polyaxonfile",
    multiple=True,
    type=click.Path(exists=True),
    help="The polyaxonfiles to update with, they should be an operation preset.",
)
@click.pass_context
@clean_outputs
def resume(ctx, project, uid, polyaxonfile):
    """Resume run.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops resume --uid=8aac02e3a62a4f0aaa257c59da5eab80
    """
    content = None
    if polyaxonfile:
        content = OperationSpecification.read(polyaxonfile, is_preset=True).to_dict(
            dump=True
        )

    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        response = polyaxon_client.resume(override_config=content)
        Printer.print_success("Run was resumed with uid {}".format(response.uuid))
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not resume run `{}`.".format(run_uuid))
        sys.exit(1)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.pass_context
@clean_outputs
def invalidate(ctx, project, uid):
    """Invalidate the run's cache state.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops invalidate
    """

    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    try:
        polyaxon_client = RunClient(
            owner=owner, project=project_name, run_uuid=run_uuid
        )
        response = polyaxon_client.invalidate()
        Printer.print_success("Run was invalidated with uid {}".format(response.uuid))
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not invalidate run `{}`.".format(run_uuid))
        sys.exit(1)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option("--watch", "-w", is_flag=True, help="Watch statuses.")
@click.pass_context
@clean_outputs
def statuses(ctx, project, uid, watch):
    """Get run or run job statuses.

    Uses /docs/core/cli/#caching

    Examples getting run statuses:

    \b
    $ polyaxon ops statuses

    \b
    $ polyaxon ops statuses -uid=8aac02e3a62a4f0aaa257c59da5eab80
    """

    def _handle_run_statuses():
        if not conditions:
            return
        Printer.print_header("Latest status:")
        latest_status = Printer.add_status_color(
            {"status": status}, status_key="status"
        )
        click.echo("{}\n".format(latest_status["status"]))

        objects = list_dicts_to_tabulate(
            [
                Printer.add_status_color(o.to_dict(), status_key="type")
                for o in conditions
            ]
        )
        if objects:
            Printer.print_header("Conditions:")
            dict_tabulate(objects, is_list_dict=True)

    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )

    client = RunClient(owner=owner, project=project_name, run_uuid=run_uuid)
    if watch:
        try:
            for status, conditions in client.watch_statuses():
                _handle_run_statuses()
        except (ApiException, HTTPError, PolyaxonClientException) as e:
            handle_cli_error(
                e, message="Could get status for run `{}`.".format(run_uuid)
            )
            sys.exit(1)
    else:
        try:
            status, conditions = client.get_statuses()
            _handle_run_statuses()
        except (ApiException, HTTPError, PolyaxonClientException) as e:
            handle_cli_error(
                e, message="Could get status for run `{}`.".format(run_uuid)
            )
            sys.exit(1)


# @ops.command()
# @click.option("--gpu", "-g", is_flag=True, help="List run GPU resources.")
# @click.pass_context
# def resources(ctx, gpu):
#     """Get run or run job resources.
#
#     Uses /docs/core/cli/#caching
#
#     Examples for getting run resources:
#
#     \b
#     $ polyaxon ops resources -uid=8aac02e3a62a4f0aaa257c59da5eab80
#
#     For GPU resources
#
#     \b
#     $ polyaxon ops resources -uid=8aac02e3a62a4f0aaa257c59da5eab80 --gpu
#     """
#
#     def get_run_resources():
#         try:
#             message_handler = Printer.gpu_resources if gpu else Printer.resources
#             PolyaxonClient().run.resources(
#                 owner, project_name, run_uuid, message_handler=message_handler
#             )
#         except (ApiException, HTTPError) as e:
#             handle_cli_error(
#                 e, message="Could not get resources for run `{}`.".format(run_uuid)
#             )
#             sys.exit(1)
#
#     owner, project_name, run_uuid = get_project_run_or_local(
#         ctx.obj.get("project"), ctx.obj.get("run_uuid"), is_cli=True,
#     )
#
#     get_run_resources()


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--follow",
    "-f",
    is_flag=True,
    default=True,
    help="Stream logs after showing past logs.",
)
@click.option(
    "--hide-time",
    "-h",
    is_flag=True,
    default=False,
    help="Whether or not to hide timestamps from the log stream.",
)
@click.option(
    "--all-containers",
    is_flag=True,
    default=False,
    help="Whether to stream logs from all containers.",
)
@click.option(
    "--all-info",
    "-a",
    is_flag=True,
    default=False,
    help="Whether to show all information including container names, pod names, and node names.",
)
@click.pass_context
@clean_outputs
def logs(ctx, project, uid, follow, hide_time, all_containers, all_info):
    """Get run or run job logs.

    Uses /docs/core/cli/#caching

    Examples for getting run logs:

    \b
    $ polyaxon run logs

    \b
    $ polyaxon ops logs -uid=8aac02e3a62a4f0aaa257c59da5eab80 -p mnist
    """

    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    client = RunClient(owner=owner, project=project_name, run_uuid=run_uuid)

    try:
        get_run_logs(
            client=client,
            hide_time=hide_time,
            all_containers=all_containers,
            all_info=all_info,
            follow=follow,
        )
    except (ApiException, HTTPError, PolyaxonClientException) as e:
        handle_cli_error(
            e,
            message="Could not get logs for run `{}`.".format(client.run_uuid),
        )
        sys.exit(1)


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--path",
    type=str,
    help="Path to download, if not provided the full run's  artifacts will downloaded.",
)
@click.option(
    "--path-to",
    type=click.Path(exists=False),
    help="The destination where to download the artifacts.",
)
@click.option(
    "--no-untar",
    is_flag=True,
    default=False,
    help="Disable the automatic untar of the downloaded the artifacts.",
)
@click.pass_context
@clean_outputs
def artifacts(ctx, project, uid, path, path_to, no_untar):
    """Download outputs/artifacts for run.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops artifacts -uid=8aac02e3a62a4f0aaa257c59da5eab80

    \b
    $ polyaxon ops artifacts -uid=8aac02e3a62a4f0aaa257c59da5eab80 path="events/only"

    \b
    $ polyaxon ops artifacts -uid=8aac02e3a62a4f0aaa257c59da5eab80 path_to="this/path"
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    try:
        client = RunClient(owner=owner, project=project_name, run_uuid=run_uuid)
        download_path = client.download_artifacts(
            path=path or "", path_to=path_to, untar=not no_untar
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e, message="Could not download outputs for run `{}`.".format(run_uuid)
        )
        sys.exit(1)
    Printer.print_success("Files downloaded: path: {}".format(download_path))


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--path-from",
    type=click.Path(exists=False),
    help="Path to use to upload. Defaults to current path.",
)
@click.option(
    "--path-to",
    type=str,
    help="The destination where to upload the artifacts. If run's root artifacts path will be used.",
)
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="If the content is a single file.",
)
@click.option(
    "--sync-failure",
    is_flag=True,
    default=False,
    help="To set the run to failed if upload fails.",
)
@click.pass_context
@clean_outputs
def upload(ctx, project, uid, path_from, path_to, is_file, sync_failure):
    """Upload runs' artifacts.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops upload -uid=8aac02e3a62a4f0aaa257c59da5eab80

    \b
    $ polyaxon ops upload -uid=8aac02e3a62a4f0aaa257c59da5eab80 path_from="path/to/upload"

    \b
    $ polyaxon ops upload -uid=8aac02e3a62a4f0aaa257c59da5eab80 path_to="path/to/upload/to"
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    try:
        client = RunClient(owner=owner, project=project_name, run_uuid=run_uuid)
        if is_file:
            response = client.upload_artifact(
                filepath=path_from, path=path_to or "", overwrite=True
            )
        else:
            response = client.upload_artifacts_dir(
                dirpath=path_from,
                path=path_to or "",
                overwrite=True,
                relative_to=path_from,
            )
    except (
        ApiException,
        HTTPError,
        PolyaxonHTTPError,
        PolyaxonShouldExitError,
        PolyaxonClientException,
    ) as e:
        handle_cli_error(
            e, message="Could not upload artifacts for run `{}`.".format(run_uuid)
        )
        sys.exit(1)

    if response.status_code == 200:
        Printer.print_success("Artifacts uploaded.")
    else:
        if sync_failure:
            client.log_failed(
                reason="OperationCli", message="Operation failed uploading artifacts."
            )
        Printer.print_error(
            "Error uploading artifacts. "
            "Status: {}. Error: {}.".format(response.status_code, response.content),
            sys_exit=True,
        )


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Automatic yes to prompts. "
    'Assume "yes" as answer to all prompts and run non-interactively.',
)
@click.option(
    "--url",
    is_flag=True,
    default=False,
    help="Print the url of the dashboard for this run.",
)
@click.pass_context
@clean_outputs
def dashboard(ctx, project, uid, yes, url):
    """Open this operation's dashboard details in browser."""
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    subpath = "{}/{}/runs/{}".format(owner, project_name, run_uuid)
    get_dashboard(
        dashboard_url=get_dashboard_url(subpath=subpath), url_only=url, yes=yes
    )


@ops.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_RUN_UID["args"], **OPTIONS_RUN_UID["kwargs"])
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Automatic yes to prompts. "
    'Assume "yes" as answer to all prompts and run non-interactively.',
)
@click.option(
    "--external",
    is_flag=True,
    default=False,
    help="Open the external service URL.",
)
@click.option(
    "--url",
    is_flag=True,
    default=False,
    help="Print the url of the dashboard or external service.",
)
@click.pass_context
@clean_outputs
def service(ctx, project, uid, yes, external, url):
    """Open the operation service in browser.

    N.B. The operation must have a run kind service, otherwise it will raise an error.

    You can open the service embedded in Polyaxon UI or using the real service URL,
    please use the `--external` flag.
    """
    owner, project_name, run_uuid = get_project_run_or_local(
        project or ctx.obj.get("project"),
        uid or ctx.obj.get("run_uuid"),
        is_cli=True,
    )
    client = RunClient(owner=owner, project=project_name, run_uuid=run_uuid)
    client.refresh_data()
    if client.run_data.kind != V1RunKind.SERVICE:
        Printer.print_warning(
            "Command expected an operation of "
            "kind `service` received kind: `{}`!".format(client.run_data.kind)
        )
        sys.exit(1)

    Printer.print_header("Waiting for running condition ...")
    client.wait_for_condition(
        statuses={V1Statuses.RUNNING} | LifeCycle.DONE_VALUES, print_status=True
    )

    client.refresh_data()
    if LifeCycle.is_done(client.run_data.status):
        Printer.print_header("The operations reached a done statuses.")
        latest_status = Printer.add_status_color(
            {"status": client.run_data.status}, status_key="status"
        )
        click.echo("{}\n".format(latest_status["status"]))

    run_url = get_dashboard_url(
        subpath="{}/{}/runs/{}/service".format(owner, project_name, run_uuid)
    )

    namespace = client.run_data.settings.namespace
    service_endpoint = SERVICES_V1
    if client.run_data.meta_info.get(META_REWRITE_PATH, False):
        service_endpoint = REWRITE_SERVICES_V1
    external_run_url = get_dashboard_url(
        base=service_endpoint,
        subpath="{}/{}/{}/runs/{}/".format(namespace, owner, project_name, run_uuid),
    )

    if url:
        Printer.print_header("The service will be available at: {}".format(run_url))
        Printer.print_header(
            "You can also view it in an external link at: {}".format(external_run_url)
        )
        sys.exit(0)
    if not yes:
        click.confirm(
            "Dashboard page will now open in your browser. Continue?",
            abort=True,
            default=True,
        )
    if external:
        click.launch(external_run_url)
        sys.exit(0)
    click.launch(run_url)


@ops.command()
@click.option(
    "--uid",
    "-uid",
    help="To single a single offline run using a its uuid.",
)
@click.option(
    "--all-runs",
    "-a",
    is_flag=True,
    default=False,
    help="To sync all runs.",
)
@click.option(
    "--no-artifacts",
    is_flag=True,
    default=False,
    help="To disable uploading artifacts.",
)
@click.option(
    "--clean",
    "-c",
    is_flag=True,
    default=False,
    help="To clean the run(s) local data after syncing.",
)
@click.option(
    "--offline-path",
    type=click.Path(exists=False),
    help="Optional path to use to where offline runs are persisted, "
    "default value is taken from the env var: `POLYAXON_OFFLINE_ROOT`.",
)
@clean_outputs
def sync(uid, all_runs, no_artifacts, clean, offline_path):
    """Syncs multiple offline runs or a single offline run if a uuid is passed.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon ops sync -a --clean

    \b
    $ polyaxon ops sync -uid=8aac02e3a62a4f0aaa257c59da5eab80 --clean

    """
    offline_path = offline_path or container_contexts.CONTEXT_OFFLINE_ROOT
    offline_path_format = "{}/{{}}".format(offline_path)

    def _sync(run_uuid: str):
        Printer.print_header(f"Syncing offline run {uid} ...")
        client = RunClient(run_uuid=run_uuid, is_offline=True)
        try:
            client.sync_offline_run(
                load_offline_run=True,
                artifacts_path=offline_path_format.format(run_uuid),
                upload_artifacts=not no_artifacts,
                clean=clean,
            )
        except (
            ApiException,
            HTTPError,
            PolyaxonHTTPError,
            PolyaxonShouldExitError,
            PolyaxonClientException,
        ) as e:
            handle_cli_error(
                e, message="Could not sync offline run `{}`.".format(run_uuid)
            )
            sys.exit(1)

    if all_runs:
        if (
            not os.path.exists(offline_path)
            or not os.path.isdir(offline_path)
            or not os.listdir(offline_path)
        ):
            Printer.print_error(
                f"Could not sync offline runs, the path `{offline_path}` "
                f"does not exist, is not a directory, or is empty."
            )
            sys.exit(1)
        for uid in os.listdir(offline_path):
            _sync(uid)
    elif uid:
        _sync(uid)
    else:
        Printer.print_error(
            f"Please provide a run uuid or pass the flag `-a/--all` to sync runs."
        )
        sys.exit(1)
